package net.knarcraft.minecraftserverlauncher.utility;

import net.knarcraft.minecraftserverlauncher.Main;
import net.knarcraft.minecraftserverlauncher.server.ServerVersionContainer;
import net.knarcraft.minecraftserverlauncher.userinterface.GUI;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.nio.file.Paths;

/**
 * The jar builder is responsible for building the newest Spigot and CraftBukkit .jar files using BuildTools
 */
public class JarBuilder {

    private final String buildDirectory;
    private final String jarDirectory;
    private final ServerVersionContainer versionContainer;
    private final GUI gui;

    /**
     * Instantiates a new jar builder
     * @param buildDirectory <p>The directory containing BuildTool files</p>
     * @param jarDirectory <p>The directory containing downloaded .jar files</p>
     * @param gui <p>The GUI to write messages to</p>
     */
    public JarBuilder(String buildDirectory, String jarDirectory, GUI gui) {
        this.buildDirectory = buildDirectory;
        this.jarDirectory = jarDirectory;
        this.gui = gui;
        this.versionContainer = ServerVersionContainer.getInstance();
    }

    /**
     * Builds the latest version of the spigot .jar file
     */
    public void buildSpigotJar() {
        downloadBuildTools();
        ProcessBuilder processBuilder = new ProcessBuilder("java", "-jar", "BuildTools.jar", "--rev",
                "latest", "--output-dir", jarDirectory);
        executeBuildProcess(processBuilder);
        gui.setStatus("Finished building spigot .jar. Moving it to the correct location");
        moveBuiltJar("spigot-", "SpigotLatest.jar");
        gui.setStatus("Finished moving spigot .jar");
    }

    /**
     * Builds the latest version of the craftbukkit .jar file
     */
    public void buildBukkitJar() {
        downloadBuildTools();
        ProcessBuilder processBuilder = new ProcessBuilder("java", "-jar", "BuildTools.jar", "--compile",
                "craftbukkit", "--rev", "latest", "--output-dir", jarDirectory);
        executeBuildProcess(processBuilder);
        gui.setStatus("Finished building craftbukkit .jar. Moving it to the correct location");
        moveBuiltJar("craftbukkit-", "BukkitLatest.jar");
        gui.setStatus("Finished moving craftbukkit .jar");
    }

    /**
     * Downloads the latest BuildTools version
     */
    public void downloadBuildTools() {
        try {
            String latestVersion = getLatestBuildToolsVersion();
            boolean exists = new File(buildDirectory + "BuildTools.jar").exists();
            boolean isUpdated = latestVersion.equals(versionContainer.getDownloadedBuildToolsVersion());

            if (!exists || !isUpdated) {
                String buildToolsURL = "https://hub.spigotmc.org/jenkins/job/BuildTools/lastSuccessfulBuild/artifact/target/BuildTools.jar";
                boolean success = CommonFunctions.downloadFile(buildToolsURL, Paths.get(buildDirectory + "BuildTools.jar"));
                if (!success) {
                    Main.getController().getGUI().setStatus("Unable to download the latest BuildTools");
                } else {
                    versionContainer.setDownloadedBuildToolsVersion(latestVersion);
                }
            }
        } catch (IOException e) {
            Main.getController().getGUI().setStatus("Unable to download the latest BuildTools");
        }
    }

    /**
     * Moves a built .jar file to its target file
     * @param searchString <p>The start of the name of the built file used to find it</p>
     * @param newName <p>The new name of the built file</p>
     */
    private void moveBuiltJar(String searchString, String newName) {
        File[] foundFiles = new File(jarDirectory).listFiles((file) -> file.isFile() && file.getName().startsWith(searchString));
        if (foundFiles != null && foundFiles.length == 1) {
            File newFileLocation = new File(jarDirectory + File.separator + newName);
            if (newFileLocation.exists()) {
                if (!newFileLocation.delete()) {
                    throw new IllegalArgumentException("Unable to delete previously built .jar");
                }
            }
            if (!foundFiles[0].renameTo(newFileLocation)) {
                throw new IllegalArgumentException("Unable to move built .jar");
            }
        }
    }

    /**
     * Starts the build process and initializes
     * @param processBuilder <p>The process builder to execute</p>
     */
    private void executeBuildProcess(ProcessBuilder processBuilder) {
        processBuilder.directory(new File(buildDirectory));
        processBuilder.redirectErrorStream(true);
        try {
            Process process = processBuilder.start();
            BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(System.out));
            BufferedReader reader = new BufferedReader(new InputStreamReader(process.getInputStream()));

            do {
                writer.write(CommonFunctions.readBufferedReader(reader));
            } while (process.isAlive());

            writer.flush();

            if (process.exitValue() == 0) {
                gui.setStatus("Jar building process finished successfully.");
            } else {
                throw new RuntimeException("Jar building process failed with exit code " + process.exitValue() + ".");
            }
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    /**
     * Gets the latest build tools version available
     * @return <p>The latest build tools version available</p>
     * @throws IOException <p>If unable to read the version file</p>
     */
    String getLatestBuildToolsVersion() throws IOException {
        String versionDocument;
        versionDocument = CommonFunctions.readRemoteFile("https://hub.spigotmc.org/jenkins/job/BuildTools/lastSuccessfulBuild/");
        return CommonFunctions.stringBetween(versionDocument, "BuildTools #", " [");
    }

}
